"""PySerials"""

from pyserials import exception, update, validate, read, write, format, compare
from pyserials.nested_dict import NestedDict
from pyserials.property_dict import PropertyDict
from pyserials.flatten import flatten
